@@ -6,6 +6,7 @@ from decimal import Decimal
 from sql import Table, Literal
 from sql.functions import Overlay, Position
 from sql.aggregate import Count
+from sql.operators import Concat
 
 from trytond.model import Workflow, ModelView, ModelSQL, fields
 from trytond.modules.company import CompanyReport
@@ -51,6 +52,7 @@ class Purchase(Workflow, ModelSQL, ModelView):
         ('draft', 'Draft'),
         ('quotation', 'Quotation'),
         ('confirmed', 'Confirmed'),
+        ('processing', 'Processing'),
         ('done', 'Done'),
         ('cancel', 'Canceled'),
     ], 'State', readonly=True, required=True)
@@ -61,9 +63,17 @@ class Purchase(Workflow, ModelSQL, ModelView):
             },
         depends=['state'])
     payment_term = fields.Many2One('account.invoice.payment_term',
-        'Payment Term', required=True, states=_STATES, depends=_DEPENDS)
+        'Payment Term', states={
+            'readonly': ~Eval('state').in_(['draft', 'quotation']),
+            'required': ~Eval('state').in_(['draft', 'quotation', 'cancel']),
+            },
+        depends=['state'])
     party = fields.Many2One('party.party', 'Party', required=True,
-        states=_STATES, select=True, depends=_DEPENDS)
+        states={
+            'readonly': ((Eval('state') != 'draft')
+                | (Eval('lines', [0]) & Eval('party'))),
+            },
+        select=True, depends=['state'])
     party_lang = fields.Function(fields.Char('Party Language'),
         'on_change_with_party_lang')
     invoice_address = fields.Many2One('party.address', 'Invoice Address',
@@ -145,8 +155,6 @@ class Purchase(Workflow, ModelSQL, ModelView):
         cls._order.insert(0, ('purchase_date', 'DESC'))
         cls._order.insert(1, ('id', 'DESC'))
         cls._error_messages.update({
-                'invoice_address_required': ('Invoice address must be '
-                    'defined for quotation of purchase "%s".'),
                 'warehouse_required': ('A warehouse must be defined for '
                     'quotation of purchase "%s".'),
                 'missing_account_payable': ('Missing "Account Payable" on '
@@ -157,8 +165,10 @@ class Purchase(Workflow, ModelSQL, ModelView):
         cls._transitions |= set((
                 ('draft', 'quotation'),
                 ('quotation', 'confirmed'),
-                ('confirmed', 'confirmed'),
-                ('done', 'confirmed'),
+                ('confirmed', 'processing'),
+                ('processing', 'processing'),
+                ('processing', 'done'),
+                ('done', 'processing'),
                 ('draft', 'cancel'),
                 ('quotation', 'cancel'),
                 ('quotation', 'draft'),
@@ -174,12 +184,20 @@ class Purchase(Workflow, ModelSQL, ModelView):
                         'tryton-go-previous'),
                     },
                 'quote': {
+                    'pre_validate': [
+                        ('purchase_date', '!=', None),
+                        ('payment_term', '!=', None),
+                        ('invoice_address', '!=', None),
+                        ],
                     'invisible': Eval('state') != 'draft',
                     'readonly': ~Eval('lines', []),
                     },
                 'confirm': {
                     'invisible': Eval('state') != 'quotation',
                     },
+                'process': {
+                    'invisible': Eval('state') != 'confirmed',
+                    },
                 'handle_invoice_exception': {
                     'invisible': ((Eval('invoice_state') != 'exception')
                         | (Eval('state') == 'cancel')),
@@ -198,6 +216,10 @@ class Purchase(Workflow, ModelSQL, ModelView):
 
     @classmethod
     def __register__(cls, module_name):
+        pool = Pool()
+        PurchaseLine = pool.get('purchase.line')
+        Move = pool.get('stock.move')
+        InvoiceLine = pool.get('account.invoice.line')
         TableHandler = backend.get('TableHandler')
         cursor = Transaction().cursor
         model_data = Table('ir_model_data')
@@ -245,6 +267,34 @@ class Purchase(Workflow, ModelSQL, ModelView):
         # Migration from 2.2: purchase_date is no more required
         table.not_null_action('purchase_date', 'remove')
 
+        # Migration from 3.2
+        # state confirmed splitted into confirmed and processing
+        if (TableHandler.table_exist(cursor, PurchaseLine._table)
+                and TableHandler.table_exist(cursor, Move._table)
+                and TableHandler.table_exist(cursor, InvoiceLine._table)):
+            purchase_line = PurchaseLine.__table__()
+            move = Move.__table__()
+            invoice_line = InvoiceLine.__table__()
+            # Wrap subquery inside an other inner subquery because MySQL syntax
+            # doesn't allow update a table and select from the same table in a
+            # subquery.
+            sub_query = sql_table.join(purchase_line,
+                condition=purchase_line.purchase == sql_table.id
+                ).join(invoice_line, 'LEFT',
+                    condition=(invoice_line.origin ==
+                        Concat(PurchaseLine.__name__, purchase_line.id))
+                    ).join(move, 'LEFT',
+                        condition=(move.origin == Concat(PurchaseLine.__name__,
+                                purchase_line.id))
+                        ).select(sql_table.id,
+                            where=((sql_table.state == 'confirmed')
+                                & ((invoice_line.id != None)
+                                    | (move.id != None))))
+            cursor.execute(*sql_table.update(
+                    columns=[sql_table.state],
+                    values=['processing'],
+                    where=sql_table.id.in_(sub_query.select(sub_query.id))))
+
         # Add index on create_date
         table = TableHandler(cursor, cls, module_name)
         table.index_action('create_date', action='add')
@@ -302,7 +352,7 @@ class Purchase(Workflow, ModelSQL, ModelView):
     def default_shipment_state():
         return 'none'
 
-    @fields.depends('party', 'payment_term')
+    @fields.depends('party', 'payment_term', 'lines')
     def on_change_party(self):
         pool = Pool()
         PaymentTerm = pool.get('account.invoice.payment_term')
@@ -312,9 +362,10 @@ class Purchase(Workflow, ModelSQL, ModelView):
         changes = {
             'invoice_address': None,
             'payment_term': None,
-            'currency': self.default_currency(),
-            'currency_digits': self.default_currency_digits(),
             }
+        if not self.lines:
+            changes['currency'] = self.default_currency()
+            changes['currency_digits'] = self.default_currency_digits()
         invoice_address = None
         payment_term = None
         if self.party:
@@ -322,19 +373,20 @@ class Purchase(Workflow, ModelSQL, ModelView):
             if self.party.supplier_payment_term:
                 payment_term = self.party.supplier_payment_term
 
-            subquery = table.select(table.currency,
-                where=table.party == self.party.id,
-                order_by=table.id,
-                limit=10)
-            cursor.execute(*subquery.select(subquery.currency,
-                    group_by=subquery.currency,
-                    order_by=Count(Literal(1)).desc))
-            row = cursor.fetchone()
-            if row:
-                currency_id, = row
-                currency = Currency(currency_id)
-                changes['currency'] = currency.id
-                changes['currency_digits'] = currency.digits
+            if not self.lines:
+                subquery = table.select(table.currency,
+                    where=table.party == self.party.id,
+                    order_by=table.id,
+                    limit=10)
+                cursor.execute(*subquery.select(subquery.currency,
+                        group_by=subquery.currency,
+                        order_by=Count(Literal(1)).desc))
+                row = cursor.fetchone()
+                if row:
+                    currency_id, = row
+                    currency = Currency(currency_id)
+                    changes['currency'] = currency.id
+                    changes['currency_digits'] = currency.digits
 
         if invoice_address:
             changes['invoice_address'] = invoice_address.id
@@ -376,6 +428,9 @@ class Purchase(Workflow, ModelSQL, ModelView):
         pool = Pool()
         Tax = pool.get('account.tax')
         Invoice = pool.get('account.invoice')
+        Configuration = pool.get('account.configuration')
+
+        config = Configuration(1)
 
         changes = {
             'untaxed_amount': Decimal('0.0'),
@@ -385,6 +440,12 @@ class Purchase(Workflow, ModelSQL, ModelView):
         if self.lines:
             context = self.get_tax_context()
             taxes = {}
+
+            def round_taxes():
+                if self.currency:
+                    for key, value in taxes.iteritems():
+                        taxes[key] = self.currency.round(value)
+
             for line in self.lines:
                 if getattr(line, 'type', 'line') != 'line':
                     continue
@@ -397,13 +458,15 @@ class Purchase(Workflow, ModelSQL, ModelView):
                         getattr(line, 'quantity', None) or 0.0)
                 for tax in tax_list:
                     key, val = Invoice._compute_tax(tax, 'in_invoice')
-                    if not key in taxes:
+                    if key not in taxes:
                         taxes[key] = val['amount']
                     else:
                         taxes[key] += val['amount']
-            if self.currency:
-                for value in taxes.itervalues():
-                    changes['tax_amount'] += self.currency.round(value)
+                if config.tax_rounding == 'line':
+                    round_taxes()
+            if config.tax_rounding == 'document':
+                round_taxes()
+            changes['tax_amount'] = sum(taxes.itervalues(), Decimal('0.0'))
         if self.currency:
             changes['untaxed_amount'] = self.currency.round(
                 changes['untaxed_amount'])
@@ -419,23 +482,34 @@ class Purchase(Workflow, ModelSQL, ModelView):
         pool = Pool()
         Tax = pool.get('account.tax')
         Invoice = pool.get('account.invoice')
+        Configuration = pool.get('account.configuration')
+
+        config = Configuration(1)
 
         context = self.get_tax_context()
         taxes = {}
+
+        def round_taxes():
+            for key, value in taxes.iteritems():
+                taxes[key] = self.currency.round(value)
+
         for line in self.lines:
             if line.type != 'line':
                 continue
             with Transaction().set_context(context):
                 tax_list = Tax.compute(line.taxes, line.unit_price,
                     line.quantity)
-            # Don't round on each line to handle rounding error
             for tax in tax_list:
                 key, val = Invoice._compute_tax(tax, 'in_invoice')
-                if not key in taxes:
+                if key not in taxes:
                     taxes[key] = val['amount']
                 else:
                     taxes[key] += val['amount']
-        return sum((self.currency.round(tax) for tax in taxes.values()), _ZERO)
+            if config.tax_rounding == 'line':
+                round_taxes()
+        if config.tax_rounding == 'document':
+            round_taxes()
+        return sum(taxes.itervalues(), _ZERO)
 
     @classmethod
     def get_amount(cls, purchases, names):
@@ -588,8 +662,6 @@ class Purchase(Workflow, ModelSQL, ModelView):
         return super(Purchase, cls).copy(purchases, default=default)
 
     def check_for_quotation(self):
-        if not self.invoice_address:
-            self.raise_user_error('invoice_address_required', (self.rec_name,))
         for line in self.lines:
             if (not line.to_location
                     and line.product
@@ -659,17 +731,16 @@ class Purchase(Workflow, ModelSQL, ModelView):
         else:
             journal = None
 
-        with Transaction().set_user(0, set_context=True):
-            return Invoice(
-                company=self.company,
-                type=invoice_type,
-                journal=journal,
-                party=self.party,
-                invoice_address=self.invoice_address,
-                currency=self.currency,
-                account=self.party.account_payable,
-                payment_term=self.payment_term,
-                )
+        return Invoice(
+            company=self.company,
+            type=invoice_type,
+            journal=journal,
+            party=self.party,
+            invoice_address=self.invoice_address,
+            currency=self.currency,
+            account=self.party.account_payable,
+            payment_term=self.payment_term,
+            )
 
     def create_invoice(self, invoice_type):
         '''
@@ -693,8 +764,7 @@ class Purchase(Workflow, ModelSQL, ModelView):
         invoice.lines = list(chain.from_iterable(invoice_lines.itervalues()))
         invoice.save()
 
-        with Transaction().set_user(0, set_context=True):
-            Invoice.update_taxes([invoice])
+        Invoice.update_taxes([invoice])
         return invoice
 
     def create_move(self, move_type):
@@ -712,12 +782,11 @@ class Purchase(Workflow, ModelSQL, ModelView):
 
     def _get_return_shipment(self):
         ShipmentInReturn = Pool().get('stock.shipment.in.return')
-        with Transaction().set_user(0, set_context=True):
-            return ShipmentInReturn(
-                company=self.company,
-                from_location=self.warehouse.storage_location,
-                to_location=self.party.supplier_location,
-                )
+        return ShipmentInReturn(
+            company=self.company,
+            from_location=self.warehouse.storage_location,
+            to_location=self.party.supplier_location,
+            )
 
     def create_return_shipment(self, return_moves):
         '''
@@ -727,8 +796,7 @@ class Purchase(Workflow, ModelSQL, ModelView):
         return_shipment = self._get_return_shipment()
         return_shipment.moves = return_moves
         return_shipment.save()
-        with Transaction().set_user(0, set_context=True):
-            ShipmentInReturn.wait([return_shipment])
+        ShipmentInReturn.wait([return_shipment])
         return return_shipment
 
     def is_done(self):
@@ -774,7 +842,11 @@ class Purchase(Workflow, ModelSQL, ModelView):
     def confirm(cls, purchases):
         cls.set_purchase_date(purchases)
         cls.store_cache(purchases)
-        cls.process(purchases)
+
+    @classmethod
+    @Workflow.transition('processing')
+    def proceed(cls, purchases):
+        pass
 
     @classmethod
     @ModelView.button_action('purchase.wizard_invoice_handle_exception')
@@ -787,8 +859,9 @@ class Purchase(Workflow, ModelSQL, ModelView):
         pass
 
     @classmethod
+    @ModelView.button
     def process(cls, purchases):
-        done = []
+        process, done = [], []
         for purchase in purchases:
             purchase.create_invoice('in_invoice')
             purchase.create_invoice('in_credit_note')
@@ -800,6 +873,10 @@ class Purchase(Workflow, ModelSQL, ModelView):
             purchase.set_shipment_state()
             if purchase.is_done():
                 done.append(purchase)
+            elif purchase.state != 'processing':
+                process.append(purchase)
+        if process:
+            cls.proceed(process)
         if done:
             cls.write(done, {
                     'state': 'done',
@@ -1118,8 +1195,7 @@ class PurchaseLine(ModelSQL, ModelView):
 
     def get_amount(self, name):
         if self.type == 'line':
-            return self.purchase.currency.round(
-                Decimal(str(self.quantity)) * self.unit_price)
+            return self.on_change_with_amount()
         elif self.type == 'subtotal':
             amount = Decimal('0.0')
             for line2 in self.purchase.lines:
@@ -1171,8 +1247,7 @@ class PurchaseLine(ModelSQL, ModelView):
         Property = pool.get('ir.property')
         InvoiceLine = pool.get('account.invoice.line')
 
-        with Transaction().set_user(0, set_context=True):
-            invoice_line = InvoiceLine()
+        invoice_line = InvoiceLine()
         invoice_line.type = self.type
         invoice_line.description = self.description
         invoice_line.note = self.note
@@ -1216,10 +1291,12 @@ class PurchaseLine(ModelSQL, ModelView):
             if old_invoice_line.id not in skip_ids:
                 quantity -= Uom.compute_qty(old_invoice_line.unit,
                         old_invoice_line.quantity, self.unit)
-        invoice_line.quantity = quantity
 
-        if invoice_line.quantity <= 0.0:
+        rounding = self.unit.rounding if self.unit else 0
+        invoice_line.quantity = Uom.round(quantity, rounding)
+        if invoice_line.quantity <= 0:
             return []
+
         invoice_line.unit = self.unit
         invoice_line.product = self.product
         invoice_line.unit_price = self.unit_price
@@ -1273,15 +1350,17 @@ class PurchaseLine(ModelSQL, ModelView):
             if move not in skip:
                 quantity -= Uom.compute_qty(move.uom, move.quantity,
                     self.unit)
-        if quantity <= 0.0:
+
+        quantity = Uom.round(quantity, self.unit.rounding)
+        if quantity <= 0:
             return
+
         if not self.purchase.party.supplier_location:
             self.raise_user_error('supplier_location_required', {
                     'purchase': self.purchase.rec_name,
                     'line': self.rec_name,
                     })
-        with Transaction().set_user(0, set_context=True):
-            move = Move()
+        move = Move()
         move.quantity = quantity
         move.uom = self.unit
         move.product = self.product
